/*
 * usage - fixlabel [-d] [-g C/H/S] [-f] [part:size[:/mountpoint]]*
 *    process disklabel output and produce a new label
 *  -d  enable debugging
 *  -g  C/H/S geometry in Cyl/Head/Sectors
 *  -f  extract geometry from fdisk output
 *  part is a-h
 *  size is in Mbytes, or + for rest of disk
 *  example:
 *  disklabel -r wd0 | fixlabel a:16:/ b:200:swap d:400:/usr e:+:/var
 *   - reads the output of disklabel and produces a script for
 *  produces:
 *   g c1078 h243 s63
 *   # p partition type start length
 *   # total sectors 16441866
 *   p 1 165 63 16441803
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <strings.h>
#include <ctype.h>

int part[8];
int offset[8];
char  *filesystem[8];
int debug;
char *prog = "??";

char c_spt[32], c_tpc[32], c_spc[32], c_c[32], c_spu[32];
char
 *p_type = " ESDI", /* type: ESDI */
 *p_disk = " wd0s1", /* disk: wd0s1 */
 *p_label = " ", /* label:  */
 *p_flags = "", /* flags: */
 *p_bps = " 512", /* bytes/sector: 512 */
 *p_spt = " 0", /* sectors/track: 63 */
 *p_tpc = " 0", /* tracks/cylinder: 243 */
 *p_spc = " 0", /* sectors/cylinder: 15309 */
 *p_c = " 0", /* cylinders: 1077 */
 *p_spu = " 0", /* sectors/unit: 16503039 */
 *p_rpm = " 3600", /* rpm: 3600 */
 *p_interleave = " 1", /* interleave: 1 */
 *p_trackskew = " 0", /* trackskew: 0 */
 *p_cylinderskew = " 0", /* cylinderskew: 0 */
 *p_headswitch = " 0		# milliseconds", /* headswitch: 0           # milliseconds */
 *p_ttot = " 0	# milliseconds", /* track-to-track seek: 0  # milliseconds */
 *p_ddata = " 0 "; /* drivedata: 0  */

struct key { char *word, **var; };
struct key keylist[] = {

{"type", &p_type },
{"disk", &p_disk},
{"label", &p_label},
{"flags", &p_flags},
{"bytes/sector", &p_bps},
{"sectors/track", &p_spt},
{"tracks/cylinder", &p_tpc},
{"sectors/cylinder", &p_spc},
{"cylinders", &p_c},
{"sectors/unit", &p_spu},
{"rpm", &p_rpm},
{"interleave", &p_interleave},
{"trackskew", &p_trackskew},
{"cylinderskew", &p_cylinderskew},
{"headswitch", &p_headswitch},
{"track-to-track seek", &p_ttot},
{"drivedata", &p_ddata},

{0,0}
};

getv( char *str )
{
	int i;
	char *s;
	if( (s = strpbrk(str,":")) ){
		*s++ = 0;
		for( i = 0; keylist[i].word; ++i ){
			if( !strcmp( str, keylist[i].word ) ){
				keylist[i].var[0] = s;
				if(debug)printf("getv('%s') = %s\n", str, s );
				break;
			}
		}
	}
}

void usage(void)
{
	fprintf(stderr,"usage - %s [-d] [-g C/H/S] [-f] [part:size[:mount]]*\n" , prog );
	fprintf(stderr,"   process disklabel output and produce a new label");
	fprintf(stderr," -d   enable debugging\n");
	fprintf(stderr," -g  C/H/S geometry in Cyl/Head/Sectors\n");
	fprintf(stderr," -f  extract geometry from fdisk output\n");
	fprintf(stderr," part is a-h\n");
	fprintf(stderr," size is in Mbytes, or + for rest of disk\n");
	fprintf(stderr," example: %s a:16:/ b:200:swap d:400:/usr e:+:/var\n", prog);
	exit(1);
}

int main( int argc, char *argv[], char *envp[] )
{
	char buf[10240], *arg, *s, *t, *fsize, *bsize, *bps, *type;
	int i, spc, c, h, sec, n, rest, noinput;
	int sz, used, left, endinfo, extract;

	if( argc > 0 ) prog = argv[0];
	if( (s = strrchr(prog,'/')) ){
		prog = s+1;
	}
	noinput = 0;
	extract = 0;
	if( argc > 0 ) prog = argv[0];
	c = h = sec = 0;
	while( (i = getopt(argc, argv, "dg:f")) != EOF ){
		switch(i){
		case 'd': debug = 1; break;
		case 'f': extract = 1; break;
		case 'g':
			noinput = 1;
			s = t = optarg;
			c = strtol(s,&t,10);
			if(t){
				s = t = t+1;
				h = strtol(s,&t,10);
			}
			if(t){
				s = t = t+1;
				sec = strtol(s,&t,10);
			}
			if(debug) printf("geom: C/H/S = %d/%d/%d\n",c,h,sec);
			break;
		default: fprintf(stderr,"bad option '%c'\n", i);
			usage(); break;
		}
	}
	for( i = optind; i < argc; ++i ){
		arg = argv[i];
		n = arg[0];
		if( isupper(n) ) n = tolower(n);
		n = n - 'a';
		if( n < 0 || n >= 8 || arg[1] != ':' ) {
			fprintf(stderr,"bad partition '%s'\n", arg);
			usage();
		}
		s = arg+2;
		if( *s == '+' ){
			part[n] = -1;
			t = s+1;
		} else {
			t = 0;
			part[n] = strtol(s,&t,10);
		}
		if( part[n] == 0 || t == 0 || (*t && *t != ':')  ){
		  fprintf(stderr,"bad partition size '%s'\n", arg);
		}
		if( t )
		filesystem[n] = t+1;
	}
	part[2] = 0;
	filesystem[2] = 0;
	for( i = 0; i < 8; ++i ){
		s = filesystem[i];
		if( s == 0 ) s = "";
		if(debug)printf("part '%c':%d, file system '%s'\n",
			i+'a', part[i], part[i] * 2048, s );
		/* convert to sectors */
		part[i] *= 2048;
	}

	if( !noinput ){
		buf[sizeof(buf)-1] = 0;
		n = read( 0, buf, sizeof(buf)-1 );
		if( n < 0 ){
			perror( "error reading stdin" );
			exit(1);
		} else if( n == sizeof(buf)-1 ){
			fprintf(stderr,"internal buffer too small!!!\n");
			usage();
		}
		buf[n] = 0;
		s = buf;
		endinfo = 0;
		for( i = 0; s; ++i ){
			t = s;
			if( (s = strchr(s,'\n')) ){
				*s++ = 0;
			}
			if(debug) printf( "DOING '%s'\n", t);
			if( extract ){
				char *v = t, *w;
				if( strncmp(t,"cylinders",9) ) continue;
				while( t ){
					v = t;
					if( (t = strpbrk(t," \n")) ) *t++ = 0;
					if( (w = strchr(v,'=')) ){
						*w++ = 0;
						if(!strncmp(v,"cyl",3)) c = atoi(w);
						if(!strncmp(v,"heads",5)) h = atoi(w);
						if(!strncmp(v,"sec",3)) sec = atoi(w);
					}
				}
				spc = h*sec;
				sz = c*spc;
				break;
			} else {
				getv(t);
			}
		}
		if( !extract ){
			c = atoi(p_c);
			h = atoi(p_tpc);
			sec = atoi(p_spc);
			sz = atoi(p_spu);
			spc = atoi(p_spc);
		}
		if(debug) printf("cyl %d, h %d, sec %d, sz %d, spc %d\n",
			c, h, sec, sz, spc );
	}
	if( noinput || extract ){
		snprintf(c_spt,sizeof(c_spt)," %d",sec); p_spt = c_spt;
		snprintf(c_tpc,sizeof(c_tpc)," %d",h); p_tpc = c_tpc;
		snprintf(c_c,sizeof(c_c)," %d",c-1); p_c = c_c;
		snprintf(c_tpc,sizeof(c_tpc)," %d",h); p_tpc = c_tpc;
		spc = sec*h;
		snprintf(c_spc,sizeof(c_spc)," %d",spc); p_spc = c_spc;
		sz = sec*h*c-sec;
		snprintf(c_spu,sizeof(c_spu)," %d",sz); p_spu = c_spu;
	}
	used = 0;
	rest = -1;
	for( i = 0; i < 8; ++i ){
		if( part[i] ){
			if( part[i] > 0 ){
				used += part[i];
			} else if( rest < 0 ){
				rest = i;
			} else usage();
		}
	}
	left = sz - used; 
	if( left < 0 ){
		fprintf( stderr, "not enough disk space for partitions\n");
		usage();
	}
	if( rest >= 0 ) part[rest] = left;
	used = 0;
	for( i = 0; i < 8; ++i ){
		if( part[i] ){
			++rest;
			offset[i] = used;
			used += part[i];
		}
	}

printf("type:%s\n", p_type );
printf("disk:%s\n", p_disk);
printf("label:%s\n", p_label);
printf("flags:%s\n", p_flags);
printf("bytes/sector:%s\n", p_bps);
printf("sectors/track:%s\n", p_spt);
printf("tracks/cylinder:%s\n", p_tpc);
printf("sectors/cylinder:%s\n", p_spc);
printf("cylinders:%s\n", p_c);
printf("sectors/unit:%s\n", p_spu);
printf("rpm:%s\n", p_rpm);
printf("interleave:%s\n", p_interleave);
printf("trackskew:%s\n", p_trackskew);
printf("cylinderskew:%s\n", p_cylinderskew);
printf("headswitch:%s\n", p_headswitch);
printf("track-to-track seek:%s\n", p_ttot);
printf("drivedata:%s\n", p_ddata);

	part['c'-'a'] = sz;
	printf(
"\n8 partitions:\n"
"#        size   offset    fstype   [fsize bsize bps/cpg]\n"
	);

	for( i = 0; i < 8; ++i ){
		if( part[i] ){
			int startc, endc, starteven, endeven;
			char line[128];

			startc = offset[i]/spc;
			starteven = ( startc * spc == offset[i] );
			endc = (part[i]+offset[i]-1)/spc;
			endeven = ((part[i]+offset[i])/spc != endc);
			snprintf( line,sizeof(line),"	# (Cyl. %4d%c- %d%c)",
				startc, starteven? ' ': '*',
				endc, endeven? ' ': '*' );
			if( debug ) printf("line '%s'\n",line);
			type = "4.2BSD";
			fsize="1024";
			bsize="8192";
			bps="   16";
			if( i == 1 ){
				type = "swap";
				fsize = "    ";
				bsize = "    ";
				bps = "  ";
			} else if (i == 2 ){
				type = "unused";
				fsize = "0";
				bsize = "0";
				bps = "     ";
			}
			printf("  %c: %8d %8d  %8s %8s %5s %s %s\n",
			i+'a',part[i],offset[i],type,fsize,bsize,bps,line );
		}
	}
	return 0;
}
/*
19: '8 partitions:'
20: '#        size   offset    fstype   [fsize bsize bps/cpg]'
21: '  a:   262144        0    4.2BSD     1024  8192    16      # (Cyl.    0 - 17*)'
22: '  b:   409600   262144      swap                           # (Cyl.   17*- 43*)'
23: '  c: 16503039        0    unused        0     0            # (Cyl.    0 - 1077*)'
24: '  e:  4096000   671744    4.2BSD     1024  8192    16      # (Cyl.   43*- 311*)'
25: '  f:  4096000  4767744    4.2BSD     1024  8192    16      # (Cyl.  311*- 578*)'
26: '  g:  4096000  8863744    4.2BSD     1024  8192    16      # (Cyl.  578*- 846*)'
27: '  h:  3543295 12959744    4.2BSD     1024  8192    16      # (Cyl.  846*- 1077*)'
*/
